from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass

import numpy as np
import pandas as pd

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.dpcr import ContainerType
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.dpcr import (
    CalculatedDataItem,
    Data,
    DataSource,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.appbio_absolute_q.constants import (
    AGGREGATION_LOOKUP,
    BRAND_NAME,
    CALCULATED_DATA_REFERENCE,
    CalculatedDataSource,
    DEVICE_TYPE,
    PLATE_WELL_COUNT,
    PRODUCT_MANUFACTURER,
    SOFTWARE_NAME,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


@dataclass
class CalculatedItem:
    identifier: str
    name: str
    value: float
    unit: str
    source: CalculatedDataSource
    source_features: list[str]

    def get_data_sources(
        self, measurement_ids: list[str], calculated_data_ids: dict[str, str]
    ) -> list[DataSource]:
        if self.source == CalculatedDataSource.CALCULATED_DATA:
            return [
                DataSource(calculated_data_ids[source_feature], source_feature)
                for source_feature in self.source_features
            ]
        else:
            return [
                DataSource(identifier, self.source_features[0])
                for identifier in measurement_ids
            ]


@dataclass
class Group:
    well_identifier: str | None
    group_identifier: str
    target_identifier: str
    calculated_data: list[CalculatedItem]
    calculated_data_ids: dict[str, str]

    @property
    def name(self) -> str:
        return self.group_identifier.split("(")[0].strip()

    @property
    def key(self) -> str:
        return f"{self.name}_{self.target_identifier}"

    def get_calculated_data_ids(self) -> dict[str, str]:
        return {
            calculated_data.name: calculated_data.identifier
            for calculated_data in self.calculated_data
        }

    @staticmethod
    def create(data: SeriesData) -> Group:
        well_identifier = data.get(str, "Well")
        aggregation_type = AGGREGATION_LOOKUP[well_identifier]

        # TODO: if aggregation type is Replicate(Average), check for required columns
        # Raise if column(s) do not exist
        calculated_data_items = [
            CalculatedItem(
                random_uuid_str(),
                reference.name,
                data[float, reference.column],
                reference.unit,
                reference.source,
                reference.source_features,
            )
            for reference in CALCULATED_DATA_REFERENCE.get(aggregation_type, [])
        ]
        calculated_data_ids = {
            calculated_data.name: calculated_data.identifier
            for calculated_data in calculated_data_items
        }
        return Group(
            well_identifier=well_identifier,
            group_identifier=data[str, "Group"],
            target_identifier=data[str, "Target"],
            calculated_data=calculated_data_items,
            calculated_data_ids=calculated_data_ids,
        )

    @staticmethod
    def create_rows(data: pd.DataFrame) -> list[Group]:
        data = data.replace(np.nan, None)[data["Name"].isna()]
        return map_rows(data, Group.create)


@dataclass
class WellItem:
    name: str
    measurement_identifier: str
    well_identifier: str
    plate_identifier: str
    group_identifier: str
    target_identifier: str
    run_identifier: str
    instrument_identifier: str
    timestamp: str
    total_partition_count: float
    reporter_dye_setting: str
    concentration: float
    positive_partition_count: float

    @property
    def group_key(self) -> str:
        return f"{self.group_identifier}_{self.target_identifier}"

    @staticmethod
    def create(data: SeriesData) -> WellItem:
        return WellItem(
            name=data[str, "Name"],
            measurement_identifier=random_uuid_str(),
            well_identifier=data[str, "Well"],
            plate_identifier=data[str, "Plate"],
            group_identifier=data[str, "Group"],
            target_identifier=data[str, "Target"],
            run_identifier=data[str, "Run"],
            instrument_identifier=data[str, "Instrument"],
            timestamp=data[str, "Date"],
            total_partition_count=round(data[float, "Total"]),
            reporter_dye_setting=data[str, "Dye"],
            concentration=data[float, "Conc. cp/uL"],
            positive_partition_count=round(data[float, "Positives"]),
        )


@dataclass
class Well:
    items: list[WellItem]

    @staticmethod
    def create_wells(data: pd.DataFrame) -> list[Well]:
        if "Name" not in data:
            msg = "Input is missing required column 'Name'."
            raise AllotropeConversionError(msg)
        data = data.dropna(subset=["Name"]).replace(np.nan, None)
        return [
            Well(map_rows(well_data, WellItem.create))
            for _, well_data in data.groupby("Well")
        ]


def create_data(data: pd.DataFrame, file_name: str) -> Data:
    wells = Well.create_wells(data)
    groups = Group.create_rows(data)

    measurement_groups = [
        MeasurementGroup(
            experimental_data_identifier=well.items[0].run_identifier,
            plate_well_count=PLATE_WELL_COUNT,
            measurements=[
                Measurement(
                    identifier=item.measurement_identifier,
                    sample_identifier=item.name,
                    location_identifier=item.well_identifier,
                    measurement_time=item.timestamp,
                    plate_identifier=item.plate_identifier,
                    target_identifier=item.target_identifier,
                    total_partition_count=item.total_partition_count,
                    concentration=item.concentration,
                    positive_partition_count=item.positive_partition_count,
                    reporter_dye_setting=item.reporter_dye_setting,
                )
                for item in well.items
            ],
        )
        for well in wells
    ]

    # Map measurement ids to group keys
    group_to_ids = defaultdict(list)
    for well in wells:
        for item in well.items:
            group_to_ids[item.group_key].append(item.measurement_identifier)

    calculated_data_documents = [
        CalculatedDataItem(
            identifier=calculated_data.identifier,
            name=calculated_data.name,
            value=calculated_data.value,
            unit=calculated_data.unit,
            data_sources=[
                DataSource(source.identifier, source.feature)
                for source in calculated_data.get_data_sources(
                    group_to_ids[group.key], group.calculated_data_ids
                )
            ],
        )
        for group in groups
        for calculated_data in group.calculated_data
    ]

    return Data(
        Metadata(
            device_identifier=wells[0].items[0].instrument_identifier,
            brand_name=BRAND_NAME,
            device_type=DEVICE_TYPE,
            container_type=ContainerType.well_plate,
            software_name=SOFTWARE_NAME,
            product_manufacturer=PRODUCT_MANUFACTURER,
            file_name=file_name,
        ),
        measurement_groups,
        calculated_data=calculated_data_documents,
    )



from __future__ import annotations

import pandas as pd

from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    ContainerType,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    NaN,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Data,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import get_key_or_error
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.thermo_fisher_qubit4 import constants
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str

EMISSION_WAVELENGTH_TO_MEASUREMENT_COLUMN = {
    "green": "Green RFU",
    "far red": "Far Red RFU",
}


def create_measurement_group(data: SeriesData) -> MeasurementGroup:
    return MeasurementGroup(
        measurement_time=data[str, "Test Date"],
        experiment_type=data.get(str, "Assay Name"),
        measurements=[
            Measurement(
                type_=MeasurementType.FLUORESCENCE,
                identifier=random_uuid_str(),
                fluorescence=data[
                    float,
                    get_key_or_error(
                        "wavelength",
                        data.get(str, "Emission", "").lower(),
                        EMISSION_WAVELENGTH_TO_MEASUREMENT_COLUMN,
                    ),
                ],
                batch_identifier=data.get(str, "Run ID"),
                sample_identifier=data[str, "Test Name"],
                sample_volume_setting=data.get(
                    float, "Sample Volume (µL)", validate=SeriesData.NOT_NAN
                ),
                excitation_wavelength_setting=data.get(str, "Excitation"),
                emission_wavelength_setting=data[str, "Emission"],
                dilution_factor_setting=data.get(
                    float, "Dilution Factor", validate=SeriesData.NOT_NAN
                ),
                original_sample_concentration=data.get(
                    float, "Original sample conc.", NaN
                ),
                original_sample_concentration_unit=data.get(
                    str, "Units_Original sample conc."
                ),
                qubit_tube_concentration=data.get(float, "Qubit® tube conc.", NaN),
                qubit_tube_concentration_units=data.get(str, "Units_Qubit® tube conc."),
                standard_1_concentration=data.get(
                    float, "Std 1 RFU", validate=SeriesData.NOT_NAN
                ),
                standard_2_concentration=data.get(
                    float, "Std 2 RFU", validate=SeriesData.NOT_NAN
                ),
                standard_3_concentration=data.get(
                    float, "Std 3 RFU", validate=SeriesData.NOT_NAN
                ),
            )
        ],
    )


def create_data(data_frame: pd.DataFrame, file_name: str) -> Data:
    return Data(
        metadata=Metadata(
            file_name=file_name,
            device_identifier=NOT_APPLICABLE,
            model_number=constants.MODEL_NUMBER,
            software_name=constants.QUBIT_SOFTWARE,
            product_manufacturer=constants.PRODUCT_MANUFACTURER,
            brand_name=constants.BRAND_NAME,
            device_type=constants.DEVICE_TYPE,
            container_type=ContainerType.tube,
        ),
        measurement_groups=map_rows(data_frame, create_measurement_group),
    )



# generated by datamodel-codegen:
#   filename:  cell-culture-analyzer.schema.json
#   timestamp: 2024-06-04T03:30:28+00:00

from __future__ import annotations

from dataclasses import dataclass

from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueCell,
    TNullableQuantityValueGramPerLiter,
    TNullableQuantityValueMicrometer,
    TNullableQuantityValueMillimeterOfMercury,
    TNullableQuantityValueMillimolePerLiter,
    TNullableQuantityValueMillionCellsPerMilliliter,
    TNullableQuantityValueMilliOsmolesPerKilogram,
    TNullableQuantityValueOpticalDensity,
    TNullableQuantityValuePercent,
    TNullableQuantityValuePH,
    TNullableQuantityValueTODO,
    TNullableQuantityValueUnitPerLiter,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDateTimeValue,
    TStringValue,
)


@dataclass(kw_only=True)
class DeviceSystemDocument:
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    device_serial_number: TStringValue | None = None


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue | None = None
    batch_identifier: TStringValue | None = None
    sample_role_type: str | None = None


@dataclass(kw_only=True)
class AnalyteDocumentItem:
    analyte_name: TStringValue
    molar_concentration: TNullableQuantityValueGramPerLiter | TNullableQuantityValueMillimolePerLiter | TNullableQuantityValueUnitPerLiter | None = (
        None
    )


@dataclass(kw_only=True)
class AnalyteAggregateDocument:
    analyte_document: list[AnalyteDocumentItem] | None = None


@dataclass(kw_only=True)
class MeasurementDocumentItem:
    sample_document: SampleDocument
    measurement_time: TDateTimeValue
    analyte_aggregate_document: AnalyteAggregateDocument
    pco2: TNullableQuantityValueMillimeterOfMercury | None = None
    co2_saturation: TNullableQuantityValuePercent | None = None
    po2: TNullableQuantityValueMillimeterOfMercury | None = None
    o2_saturation: TNullableQuantityValuePercent | None = None
    optical_density: TNullableQuantityValueOpticalDensity | None = None
    pH: TNullableQuantityValuePH | None = None
    osmolality: TNullableQuantityValueMilliOsmolesPerKilogram | None = None
    viability__cell_counter_: TNullableQuantityValuePercent | None = None
    total_cell_density__cell_counter_: TNullableQuantityValueMillionCellsPerMilliliter | None = (
        None
    )
    viable_cell_density__cell_counter_: TNullableQuantityValueMillionCellsPerMilliliter | None = (
        None
    )
    average_live_cell_diameter__cell_counter_: TNullableQuantityValueMicrometer | None = (
        None
    )
    average_total_cell_diameter__cell_counter_: TNullableQuantityValueMicrometer | None = (
        None
    )
    total_cell_diameter_distribution__cell_counter_: TNullableQuantityValueTODO | None = (
        None
    )
    viable_cell_count__cell_counter_: TNullableQuantityValueCell | None = None
    total_cell_count__cell_counter_: TNullableQuantityValueCell | None = None


@dataclass(kw_only=True)
class MeasurementAggregateDocument:
    measurement_identifier: TStringValue
    analyst: TStringValue
    device_system_document: DeviceSystemDocument
    measurement_document: list[MeasurementDocumentItem]
    data_processing_time: TDateTimeValue | None = None


@dataclass(kw_only=True)
class Model:
    manifest: str = "http://purl.allotrope.org/manifests/cell-culture-analyzer/BENCHLING/2023/09/cell-culture-analyzer.manifest"
    measurement_aggregate_document: MeasurementAggregateDocument | None = None



from allotropy.allotrope.models.adm.solution_analyzer.rec._2024._03.solution_analyzer import (
    AnalyteAggregateDocument,
    AnalyteDocument,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocument,
    Model,
    SampleDocument,
    SolutionAnalyzerAggregateDocument,
    SolutionAnalyzerDocumentItem,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueGramPerLiter,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMilliliterPerLiter,
    TQuantityValueMillimolePerLiter,
)
from allotropy.allotrope.models.shared.definitions.definitions import NaN
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.roche_cedex_bioht.constants import (
    OPTICAL_DENSITY,
    SOLUTION_ANALYZER,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_reader import (
    RocheCedexBiohtReader,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_structure import (
    Data,
    Measurement,
    Sample,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser


class RocheCedexBiohtParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Roche Cedex BioHT"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        contents = named_file_contents.contents
        reader = RocheCedexBiohtReader(contents)
        return self._get_model(
            Data.create(reader), named_file_contents.original_file_name
        )

    def _get_model(self, data: Data, file_name: str) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/solution-analyzer/REC/2024/03/solution-analyzer.manifest",
            solution_analyzer_aggregate_document=SolutionAnalyzerAggregateDocument(
                solution_analyzer_document=self._get_solution_analyzer_document(data),
                device_system_document=DeviceSystemDocument(
                    model_number=data.title.model_number,
                    equipment_serial_number=data.title.device_serial_number,
                    device_identifier=NOT_APPLICABLE,
                ),
                data_system_document=DataSystemDocument(
                    file_name=file_name,
                    UNC_path="",
                    software_name=data.title.model_number,
                    software_version=data.title.software_version,
                    ASM_converter_name=self.get_asm_converter_name(),
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
            ),
        )

    def _get_solution_analyzer_document(
        self, data: Data
    ) -> list[SolutionAnalyzerDocumentItem]:
        solution_analyzer_document = []
        for sample in data.samples:
            if docs := self._get_measurement_document(sample):
                solution_analyzer_document.append(
                    SolutionAnalyzerDocumentItem(
                        measurement_aggregate_document=MeasurementAggregateDocument(
                            measurement_document=docs,
                            data_processing_time=data.title.data_processing_time,
                        ),
                        analyst=data.title.analyst,
                    )
                )

        return solution_analyzer_document

    def _get_measurement_document(self, sample: Sample) -> list[MeasurementDocument]:
        measurement_document = []

        for measurement_time, measurements in sample.measurements.items():
            if doc := self._create_sample_measurement(
                sample,
                measurement_time,
                {k: v for k, v in measurements.items() if k == OPTICAL_DENSITY},
            ):
                measurement_document.append(doc)
            if doc := self._create_sample_measurement(
                sample,
                measurement_time,
                {k: v for k, v in measurements.items() if k != OPTICAL_DENSITY},
            ):
                measurement_document.append(doc)

        return measurement_document

    def _create_analyte_document(
        self, measurement: Measurement
    ) -> tuple[AnalyteDocument, str | None]:
        value = (
            measurement.concentration_value
            if isinstance(measurement.concentration_value, float)
            else -1
        )
        analyte = AnalyteDocument(
            analyte_name=measurement.name,
            molar_concentration=TQuantityValueMillimolePerLiter(value=value),
        )
        error = measurement.error

        if measurement.unit == "g/L":
            analyte = AnalyteDocument(
                analyte_name=measurement.name,
                mass_concentration=TQuantityValueGramPerLiter(value=value),
            )
        elif measurement.unit == "mL/L":
            analyte = AnalyteDocument(
                analyte_name=measurement.name,
                volume_concentration=TQuantityValueMilliliterPerLiter(value=value),
            )
        elif measurement.unit == "mmol/L":
            analyte = AnalyteDocument(
                analyte_name=measurement.name,
                molar_concentration=TQuantityValueMillimolePerLiter(value=value),
            )
        elif measurement.unit == "U/L":
            if measurement.name == "ldh":
                analyte = AnalyteDocument(
                    analyte_name=measurement.name,
                    molar_concentration=TQuantityValueMillimolePerLiter(
                        value=value * 0.0167 if value > 0 else value
                    ),
                )
            else:
                error = f"Invalid unit for {measurement.name}: {measurement.unit}"
        else:
            error = f"Invalid unit for analyte: {measurement.unit}, value values are: g/L, mL/L, mmol/L"

        return analyte, error

    def _create_sample_measurement(
        self,
        sample: Sample,
        measurement_time: str,
        measurements: dict[str, Measurement],
    ) -> MeasurementDocument | None:
        if not measurements:
            return None

        measurement_document = MeasurementDocument(
            measurement_identifier=random_uuid_str(),
            measurement_time=self._get_date_time(measurement_time),
            sample_document=SampleDocument(
                sample_identifier=sample.name,
                batch_identifier=sample.batch,
            ),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(device_type=SOLUTION_ANALYZER)
                ]
            ),
        )

        errors: list[ErrorDocumentItem] = []
        if OPTICAL_DENSITY in measurements:
            measurement = measurements[OPTICAL_DENSITY]
            value = (
                measurement.concentration_value
                if measurement.concentration_value is not NaN
                else -1
            )
            measurement_document.absorbance = TQuantityValueMilliAbsorbanceUnit(
                value=value
            )

            # TODO: always add value, add error to error document
            if measurement.concentration_value is NaN:
                return None
        else:
            analytes = []
            for name in sorted(measurements):
                analyte, error = self._create_analyte_document(measurements[name])
                # TODO: always write analyte, add error to error document
                if not error:
                    analytes.append(analyte)

            if analytes:
                measurement_document.analyte_aggregate_document = (
                    AnalyteAggregateDocument(analyte_document=analytes)
                )
            else:
                return None

        if errors:
            measurement_document.error_aggregate_document = ErrorAggregateDocument(
                error_document=errors
            )

        return measurement_document



DEVICE_TYPE = "plate reader"
# The epoch is a place holder for measurement time since is not optional but
# the input file does not include a time that can be used.
EPOCH = "1970-01-01T00:00:00-00:00"
REDUCED = "Reduced"



from collections.abc import Callable, Iterable
from typing import TypeVar

T = TypeVar("T")
S = TypeVar("S")


def get_first_not_none(func: Callable[[T], S | None], args: Iterable[T]) -> S | None:
    for arg in args:
        ret = func(arg)
        if ret is not None:
            return ret
    return None



#!/usr/bin/env python

import argparse
import json
from typing import Any

from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    DataSource,
    Referenceable,
)
from allotropy.parsers.utils.calculated_data_documents.visualization import (
    visualize_graph,
)


class Parser:
    def __init__(self, asm_dict: dict[str, Any]):
        self.asm_dict = asm_dict
        self.measurement_docs = self.get_measurement_docs()
        self.calc_docs = self.get_calc_docs()

    def get_measurement_docs(self) -> dict[str, Any]:
        measurement_agg_doc = "measurement aggregate document"
        measurement_document = "measurement document"
        return {
            measurement_doc["measurement identifier"]: measurement_doc
            for qpcr_doc in self.asm_dict["qPCR aggregate document"]["qPCR document"]
            for measurement_doc in qpcr_doc[measurement_agg_doc][measurement_document]
        }

    def get_calc_docs(self) -> dict[str, Any]:
        qpcr_doc = self.asm_dict["qPCR aggregate document"]
        calculated_agg_document = qpcr_doc["calculated data aggregate document"]
        return {
            calc_doc["calculated data identifier"]: calc_doc
            for calc_doc in calculated_agg_document["calculated data document"]
        }

    def build_data_source(self, data_source: dict[str, Any]) -> DataSource:
        data_source_id = data_source["data source identifier"]

        if measurement_doc := self.measurement_docs.get(data_source_id):
            return DataSource(
                feature=data_source["data source feature"],
                reference=Referenceable(uuid=measurement_doc["measurement identifier"]),
            )
        elif calc_doc := self.calc_docs.get(data_source_id):
            return DataSource(
                feature=calc_doc["calculated data name"],
                reference=self.build_calc_doc(calc_doc),
            )

        msg = f"Unable to find data source '{data_source_id}'"
        raise ValueError(msg)

    def build_calc_doc(self, calc_doc: dict[str, Any]) -> CalculatedDocument:
        data_source_agg = calc_doc["data source aggregate document"]
        data_sources = data_source_agg["data source document"]
        return CalculatedDocument(
            uuid=calc_doc["calculated data identifier"],
            name=calc_doc["calculated data name"],
            value=calc_doc["calculated datum"]["value"],
            data_sources=[self.build_data_source(source) for source in data_sources],
        )

    def parse(self) -> dict[str, CalculatedDocument]:
        calc_docs = [
            self.build_calc_doc(calc_doc) for calc_doc in self.calc_docs.values()
        ]
        return {calc_doc.uuid: calc_doc for calc_doc in calc_docs}


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="visualization",
        description="Graph visualization for calculated data document structure",
    )
    parser.add_argument(
        "input_file",
        type=str,
        help="asm json file to visualize",
    )
    parser.add_argument(
        "calculated_document_name",
        type=str,
        help="name of calculated data document to visualize",
    )
    return parser.parse_args()


def read_json_file(path: str) -> dict[str, Any]:
    with open(path) as f:
        data: dict[str, Any] = json.load(f)
    return data


def main() -> None:
    args = parse_args()
    data = read_json_file(args.input_file)
    calc_docs = Parser(data).parse()
    visualize_graph(calc_docs, args.calculated_document_name)


if __name__ == "__main__":
    main()



from __future__ import annotations

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ContainerType,
    DataSystemDocument,
    DeviceSystemDocument,
    LuminescencePointDetectionDeviceControlAggregateDocument,
    LuminescencePointDetectionDeviceControlDocumentItem,
    LuminescencePointDetectionMeasurementDocumentItems,
    MeasurementAggregateDocument,
    Model,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueNumber,
    TQuantityValueRelativeLightUnit,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.methodical_mind.methodical_mind_structure import (
    CombinedData,
    PlateData,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser

LUMINESCENCE = "luminescence"
LUMINESCENCE_DETECTOR = "luminescence detector"


class MethodicalMindParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Methodical Mind"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        file_lines = read_to_lines(named_file_contents)
        reader = CsvReader(file_lines)
        combined_data = CombinedData.create(reader)
        return self._get_model(combined_data)

    def _get_model(self, combined_data: CombinedData) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                device_system_document=self._get_device_system_document(combined_data),
                data_system_document=self._get_data_system_document(combined_data),
                plate_reader_document=self._get_plate_reader_document(
                    combined_data.plate_doc_info
                ),
            ),
        )

    def _get_device_system_document(
        self, combined_data: CombinedData
    ) -> DeviceSystemDocument:
        return DeviceSystemDocument(
            device_identifier=NOT_APPLICABLE,
            model_number=combined_data.model,
            equipment_serial_number=combined_data.serial_number,
        )

    def _get_data_system_document(
        self, combined_data: CombinedData
    ) -> DataSystemDocument:
        return DataSystemDocument(
            file_name=combined_data.file_name.rsplit("\\", 1)[-1],
            UNC_path=combined_data.file_name,
            software_name=combined_data.version,
            software_version=combined_data.version,
            ASM_converter_name=self.get_asm_converter_name(),
            ASM_converter_version=ASM_CONVERTER_VERSION,
        )

    def _get_plate_reader_document(
        self, plate_docs: list[PlateData]
    ) -> list[PlateReaderDocumentItem]:
        plate_reader_docs = []
        for plate in plate_docs:
            for well in plate.well_data:
                plate_reader_doc = PlateReaderDocumentItem(
                    analyst=plate.analyst,
                    measurement_aggregate_document=MeasurementAggregateDocument(
                        measurement_time=self._get_date_time(plate.measurement_time),
                        container_type=ContainerType.well_plate,
                        plate_well_count=TQuantityValueNumber(
                            value=plate.plate_well_count
                        ),
                        measurement_document=[
                            LuminescencePointDetectionMeasurementDocumentItems(
                                measurement_identifier=random_uuid_str(),
                                luminescence=TQuantityValueRelativeLightUnit(
                                    value=well.luminescence
                                ),
                                sample_document=SampleDocument(
                                    sample_identifier=well.sample_identifier,
                                    location_identifier=well.location_identifier,
                                    well_plate_identifier=plate.well_plate_id,
                                ),
                                device_control_aggregate_document=LuminescencePointDetectionDeviceControlAggregateDocument(
                                    device_control_document=[
                                        LuminescencePointDetectionDeviceControlDocumentItem(
                                            device_type=LUMINESCENCE_DETECTOR,
                                            detection_type=LUMINESCENCE,
                                        )
                                    ]
                                ),
                            )
                        ],
                    ),
                )
                plate_reader_docs.append(plate_reader_doc)
        return plate_reader_docs



from __future__ import annotations

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.dpcr import (
    Data,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.exceptions import get_key_or_error
from allotropy.parsers.qiacuity_dpcr.constants import (
    BRAND_NAME,
    DEVICE_IDENTIFIER,
    DEVICE_TYPE,
    EPOCH,
    PRODUCT_MANUFACTURER,
    SAMPLE_ROLE_TYPE_MAPPING,
    SOFTWARE_NAME,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def _create_measurements(data: SeriesData) -> Measurement:
    sample_role_type = data.get(str, "Type")
    # TODO: When the sample role type model is updated in this repo, we should update this
    # Map sample role types to valid sample role types from ASM
    if sample_role_type is not None:
        sample_role_type = get_key_or_error(
            "sample type", sample_role_type, SAMPLE_ROLE_TYPE_MAPPING
        )

    return Measurement(
        identifier=random_uuid_str(),
        measurement_time=EPOCH,
        sample_identifier=data[str, "Sample/NTC/Control"],
        sample_role_type=sample_role_type,
        location_identifier=data[str, "Well Name"],
        plate_identifier=data.get(str, "Plate ID"),
        target_identifier=data[str, "Target"],
        total_partition_count=data[int, "Partitions (valid)"],
        concentration=data[float, "Concentration (copies/µL)"],
        positive_partition_count=data[int, "Partitions (positive)"],
        negative_partition_count=data.get(int, "Partitions (negative)"),
        flourescence_intensity_threshold_setting=data.get(float, "Threshold"),
    )


def create_data(data: pd.DataFrame, file_name: str) -> Data:
    return Data(
        Metadata(
            device_identifier=DEVICE_IDENTIFIER,
            brand_name=BRAND_NAME,
            device_type=DEVICE_TYPE,
            software_name=SOFTWARE_NAME,
            product_manufacturer=PRODUCT_MANUFACTURER,
            file_name=file_name,
        ),
        measurement_groups=[
            MeasurementGroup(
                measurements=map_rows(data, _create_measurements),
                # TODO: Hardcoded plate well count to 0 since it's a required field
                #  ASM will be modified to optional in future version
                plate_well_count=0,
            )
        ],
    )



import re
from typing import Any

from allotropy.allotrope.models.adm.light_obscuration.benchling._2023._12.light_obscuration import (
    CalculatedDataDocumentItem,
    DataProcessingDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceDocumentItem,
    DeviceSystemDocument,
    DistributionAggregateDocument,
    DistributionDocumentItem,
    DistributionItem,
    LightObscurationAggregateDocument,
    LightObscurationDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    TCalculatedDataAggregateDocument,
    TDataSourceAggregateDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCountsPerMilliliter,
    TQuantityValueMicrometer,
    TQuantityValueMilliliter,
    TQuantityValueUnitless,
)
from allotropy.allotrope.pandas_util import read_excel
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_pharmspec.beckman_pharmspec_structure import (
    Distribution,
    PharmSpecData,
)
from allotropy.parsers.beckman_pharmspec.constants import PHARMSPEC_SOFTWARE_NAME
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser

# This map is used to coerce the column names coming in the raw data
# into names of the allotrope properties.


PROPERTY_LOOKUP = {
    "particle_size": TQuantityValueMicrometer,
    "cumulative_count": TQuantityValueUnitless,
    "cumulative_particle_density": TQuantityValueCountsPerMilliliter,
    "differential_particle_density": TQuantityValueCountsPerMilliliter,
    "differential_count": TQuantityValueUnitless,
}


def get_property_from_sample(property_name: str, value: Any) -> Any:
    return PROPERTY_LOOKUP[property_name](value=value)


class PharmSpecParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Beckman PharmSpec"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        df = read_excel(named_file_contents.contents, header=None, engine="calamine")
        data = PharmSpecData.create(df)
        return self._setup_model(data, named_file_contents.original_file_name)

    def _create_distribution_document_items(
        self, distribution: Distribution
    ) -> list[DistributionDocumentItem]:
        """Create the distribution document. First, we create the actual distribution, which itself
        contains a list of DistributionDocumentItem objects. The DistributionDocumentItem objects represent
        the values from the rows of the incoming dataframe.

        :param distribution: the Distribution object
        :return: The DistributionDocument
        """
        items = []
        for row in distribution.data:
            item = {}
            for key in PROPERTY_LOOKUP:
                prop = row.get_property(key)
                if prop is not None:
                    item[key] = get_property_from_sample(key, prop.value)
            item["distribution_identifier"] = row.distribution_row_id
            items.append(DistributionItem(**item))
        return [DistributionDocumentItem(distribution=items)]

    def _create_calculated_document_items(
        self, data: PharmSpecData
    ) -> list[CalculatedDataDocumentItem]:
        calcs = [x for x in data.distributions if x.is_calculated]
        sources = [x for x in data.distributions if not x.is_calculated]
        items = []
        for calc in calcs:
            for row in calc.data:
                for prop in PROPERTY_LOOKUP:
                    distribution_property = row.get_property(prop)
                    if distribution_property:
                        source_rows = [
                            x
                            for source in sources
                            for x in source.data
                            if row.matches_property_value("particle_size", x)
                        ]
                        items.append(
                            CalculatedDataDocumentItem(
                                calculated_data_identifier=distribution_property.distribution_property_id,
                                calculated_data_name=f"{calc.name}_{prop}".lower(),
                                calculated_result=get_property_from_sample(
                                    prop, distribution_property.value
                                ),
                                data_source_aggregate_document=TDataSourceAggregateDocument(
                                    data_source_document=[
                                        DataSourceDocumentItem(
                                            data_source_identifier=f"{x.distribution_row_id}",  # will be replaced by distribution id
                                            data_source_feature=prop.replace("_", " "),
                                        )
                                        for x in source_rows
                                    ]
                                ),
                            )
                        )
        return items

    def _get_software_version_report_string(self, report_string: str) -> str:
        match = re.search(r"v(\d+(?:\.\d+)?(?:\.\d+)?)", report_string)
        if match:
            return match.group(1)
        return "Unknown"

    def _create_calculated_data_aggregate_document(
        self, data: PharmSpecData
    ) -> TCalculatedDataAggregateDocument:
        return TCalculatedDataAggregateDocument(
            calculated_data_document=self._create_calculated_document_items(data)
        )

    def _create_measurement_document_items(
        self, data: PharmSpecData
    ) -> list[MeasurementDocumentItem]:
        items = []
        for d in [x for x in data.distributions if not x.is_calculated]:
            items.append(
                MeasurementDocumentItem(
                    measurement_identifier=d.name,
                    measurement_time=data.metadata.measurement_time,
                    device_control_aggregate_document=DeviceControlAggregateDocument(
                        device_control_document=[
                            DeviceControlDocumentItem(
                                flush_volume_setting=TQuantityValueMilliliter(
                                    value=data.metadata.flush_volume_setting
                                ),
                                detector_view_volume=TQuantityValueMilliliter(
                                    value=data.metadata.detector_view_volume
                                ),
                                repetition_setting=data.metadata.repetition_setting,
                                sample_volume_setting=TQuantityValueMilliliter(
                                    value=data.metadata.sample_volume_setting,
                                ),
                            )
                        ]
                    ),
                    sample_document=SampleDocument(
                        sample_identifier=data.metadata.sample_identifier,
                    ),
                    processed_data_aggregate_document=ProcessedDataAggregateDocument(
                        processed_data_document=[
                            ProcessedDataDocumentItem(
                                data_processing_document=DataProcessingDocument(
                                    dilution_factor_setting=TQuantityValueUnitless(
                                        value=data.metadata.dilution_factor_setting,
                                    )
                                ),
                                distribution_aggregate_document=DistributionAggregateDocument(
                                    distribution_document=self._create_distribution_document_items(
                                        d
                                    )
                                ),
                            )
                        ]
                    ),
                )
            )

        return items

    def _create_model(
        self,
        data: PharmSpecData,
        calc_agg_doc: TCalculatedDataAggregateDocument | None,
        measurement_doc_items: list[MeasurementDocumentItem],
        file_name: str,
    ) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/light-obscuration/BENCHLING/2023/12/light-obscuration.manifest",
            light_obscuration_aggregate_document=LightObscurationAggregateDocument(
                light_obscuration_document=[
                    LightObscurationDocumentItem(
                        measurement_aggregate_document=MeasurementAggregateDocument(
                            analyst=data.metadata.analyst,
                            submitter=data.metadata.submitter,
                            measurement_document=measurement_doc_items,
                        )
                    )
                ],
                data_system_document=DataSystemDocument(
                    file_name=file_name,
                    software_name=PHARMSPEC_SOFTWARE_NAME,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.get_asm_converter_name(),
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                device_system_document=DeviceSystemDocument(
                    equipment_serial_number=data.metadata.equipment_serial_number,
                    device_document=[
                        DeviceDocumentItem(
                            detector_identifier=data.metadata.detector_identifier,
                            detector_model_number=data.metadata.detector_model_number,
                        )
                    ],
                ),
                calculated_data_aggregate_document=calc_agg_doc,
            ),
        )

    def _setup_model(self, data: PharmSpecData, file_name: str) -> Model:
        """Build the Model

        :param df: the raw dataframe
        :return: the model
        """
        measurement_doc_items = self._create_measurement_document_items(data)
        calc_agg_doc = self._create_calculated_data_aggregate_document(data)
        return self._create_model(data, calc_agg_doc, measurement_doc_items, file_name)



from dataclasses import dataclass

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.fluorescence.benchling._2023._09.fluorescence import (
    ContainerType,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
)
from allotropy.allotrope.models.shared.components.plate_reader import (
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueNumber,
)
from allotropy.allotrope.models.shared.definitions.definitions import TDateTimeValue
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.example_weyland_yutani.example_weyland_yutani_structure import (
    Data,
)
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser


@dataclass(kw_only=True)
class MyCustomInfoDoc:
    extra_information: str
    extra_value: TQuantityValueNumber


class ExampleWeylandYutaniParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Example Weyland Yutani"

    @property
    def release_state(self) -> ReleaseState:
        # Example parser should not be used.
        return ReleaseState.WORKING_DRAFT

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        lines = read_to_lines(named_file_contents)
        reader = CsvReader(lines)
        return self._get_model(Data.create(reader))

    def _get_model(self, data: Data) -> Model:
        if data.number_of_wells is None:
            msg = "Unable to determine the number of wells in the plate."
            raise AllotropeConversionError(msg)

        custom_info = {
            "custom value": {"value": 10, "unit": "mL"},
            "custom metadata": "Some extra piece of info",
        }
        return Model(
            measurement_aggregate_document=add_custom_information_document(
                MeasurementAggregateDocument(
                    measurement_identifier=random_uuid_str(),
                    measurement_time=self._get_measurement_time(data),
                    analytical_method_identifier=data.basic_assay_info.protocol_id,
                    experimental_data_identifier=data.basic_assay_info.assay_id,
                    container_type=ContainerType.well_plate,
                    plate_well_count=TQuantityValueNumber(value=data.number_of_wells),
                    device_system_document=add_custom_information_document(
                        DeviceSystemDocument(
                            model_number=data.instrument.serial_number,
                            device_identifier=data.instrument.nickname,
                        ),
                        MyCustomInfoDoc(
                            extra_information="My extra information",
                            extra_value=TQuantityValueNumber(value=100),
                        ),
                    ),
                    measurement_document=self._get_measurement_document(data),
                ),
                custom_info,
            )
        )

    # TODO: extract and return actual measurement time
    def _get_measurement_time(self, data: Data) -> TDateTimeValue:  # noqa: ARG002
        return self._get_date_time("2022-12-31")

    def _get_measurement_document(self, data: Data) -> list[MeasurementDocumentItem]:
        device_control_aggregate_document = (
            self._get_device_control_aggregate_document()
        )
        return [
            MeasurementDocumentItem(
                sample_document=SampleDocument(
                    well_location_identifier=f"{result.col}{result.row}"
                ),
                device_control_aggregate_document=device_control_aggregate_document,
                processed_data_aggregate_document=ProcessedDataAggregateDocument(
                    processed_data_document=[
                        ProcessedDataDocumentItem(
                            processed_data=result.value,
                            data_processing_description="processed data",
                        ),
                    ]
                ),
            )
            for result in data.plates[0].results
        ]

    def _get_device_control_aggregate_document(self) -> DeviceControlAggregateDocument:
        return DeviceControlAggregateDocument(
            device_control_document=[
                DeviceControlDocumentItem(
                    device_type="fluorescence detector",
                )
            ]
        )



from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import Model
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.qpcr import Mapper
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_data_creator import (
    create_data,
)
from allotropy.parsers.lines_reader import LinesReader, read_to_lines
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AppBioQuantStudioParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "AppBio QuantStudio RT-PCR"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        lines = read_to_lines(named_file_contents)
        reader = LinesReader(lines)
        data = create_data(reader, named_file_contents.original_file_name)
        return self._get_mapper(Mapper).map_model(data)



"""Reader file for ThermoFisher Qubit 4 Adapter"""

import pandas as pd

from allotropy.allotrope.pandas_util import read_csv, read_excel
from allotropy.constants import DEFAULT_ENCODING
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.thermo_fisher_qubit4 import constants


class ThermoFisherQubit4Reader:
    """
    A reader class for parsing ThermoFisher Qubit 4 data files into pandas DataFrames.

    Methods:
    read(named_file_contents: NamedFileContents) -> pd.DataFrame:
        Reads the content of the provided named file and returns it as a pandas DataFrame.
    """

    @classmethod
    def read(cls, named_file_contents: NamedFileContents) -> pd.DataFrame:
        """
        Reads the content of the provided named file and returns it as a pandas DataFrame.
        Changes the "Units" column to "Units_{previous_column}" for better understanding

        Parameters:
        named_file_contents (NamedFileContents): The named file contents to read.
            It includes the original file name, file contents, and encoding.

        Returns:
        pd.DataFrame: The content of the file as a pandas DataFrame.

        Raises:
        AllotropeConversionError: If the file format is not supported.
        """
        if named_file_contents.original_file_name.endswith(".xlsx"):
            dataframe = read_excel(named_file_contents.contents.name)
        elif named_file_contents.original_file_name.endswith(".csv"):
            dataframe = read_csv(
                named_file_contents.contents,
                index_col=False,
                encoding=DEFAULT_ENCODING,
            )
        else:
            message = f"{constants.UNSUPPORTED_FILE_FORMAT_ERROR} '{named_file_contents.original_file_name}'"
            raise AllotropeConversionError(message)
        columns = dataframe.columns.tolist()
        new_columns = [
            f"Units_{columns[i - 1]}" if "Units" in col else col
            for i, col in enumerate(columns)
        ]
        dataframe.columns = pd.Index(new_columns)
        return dataframe



from allotropy.__about__ import __version__

ASM_CONVERTER_NAME = "allotropy"
ASM_CONVERTER_VERSION = __version__

# Value to use for encoding str params when we want to use chardet to detect the encoding.
CHARDET_ENCODING = "chardet"

DEFAULT_ENCODING = "UTF-8"



from collections import defaultdict

from allotropy.parsers.appbio_quantstudio.views import View
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_structure import (
    SAMPLE_ROLE_TYPES_MAP,
    WellItem,
)
from allotropy.parsers.utils.values import assert_not_none


class SampleView(View[WellItem]):
    def sort_elements(self, well_items: list[WellItem]) -> dict[str, list[WellItem]]:
        items = defaultdict(list)
        for well_item in well_items:
            items[well_item.sample_identifier].append(well_item)
        return dict(items)


class TargetView(View[WellItem]):
    def sort_elements(self, well_items: list[WellItem]) -> dict[str, list[WellItem]]:
        items = defaultdict(list)
        for well_item in well_items:
            items[well_item.target_dna_description].append(well_item)
        return dict(items)


class TargetRoleView(View[WellItem]):
    def sort_elements(self, well_items: list[WellItem]) -> dict[str, list[WellItem]]:
        standard_sample_role_type = assert_not_none(
            SAMPLE_ROLE_TYPES_MAP.get("STANDARD"),
            msg="Unable to get sample role type map for STANDARD value.",
        )

        items = defaultdict(list)
        for well_item in well_items:
            if well_item.sample_role_type == standard_sample_role_type:
                items[well_item.target_dna_description].append(well_item)
        return dict(items)



ASM_CONVERTER_NAME = "allotropy_biorad_bioplex_manager"
WELLS_TAG = "Wells"
ANALYTE_NAME = "AnalyteName"
MW_ANALYTES = "MWAnalytes"
RP1_GAIN = "RP1Gain"
LABEL = "Label"
RUN_CONDITIONS = "RunConditions"
MACHINE_INFO = "MachineInfo"
DOC_LOCATION_TAG = "NativeDocumentLocation"
DESCRIPTION_TAG = "Description"
PLATE_DIMENSIONS_TAG = "PlateDimensions"
TOTAL_WELLS_ATTRIB = "TotalWells"
VERSION_ATTRIB = "BioPlexManagerVersion"
REGION_COUNT = "RegionCount"
SAMPLES = "Samples"
BEAD_MAP = "BeadMap"
MEDIAN = "Median"
MEMBER_WELLS = "MemberWells"
ROW_NUMBER = "RowNo"
COLUMN_NUMBER = "ColNo"
SERIAL_NUMBER = "SerialNumber"
WELL_NUMBER = "WellNumber"
WELL_NO = "WellNo"
REGION_NUMBER = "RegionNumber"
REGIONS_OF_INTEREST = "RegionsOfInterest"
STOP_READING_CRITERIA = "StopReadingCriteria"
READING = "Reading"
RUN_SETTINGS = "RunSettings"
SAMPLE_VOLUME = "SampleVolume"
CODE = "Code"
DILUTION = "Dilution"
BEAD_COUNT = "BeadCount"
TOTAL_EVENTS = "TotalEvents"
ACQ_TIME = "AcquisitionTime"
BEAD_REGIONS = "BeadRegions"
PRODUCT_MANUFACTURER = "Bio-Rad"
SOFTWARE_NAME = "“Bio-Plex Manager"
DEVICE_TYPE = "multi analyte profiling analyzer"
USER = "User"
CONTAINER_TYPE = "well plate"

ROW_NAMES = "ABCDEFGH"

ERROR_MAPPING = {
    "1": "Low bead number",
    "2": "Aggregated beads",
    "3": "Classify efficiency",
    "4": "Region selection",
    "5": "Platform temperature",
}



from __future__ import annotations

import math
import re
from typing import Any, TypeVar

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import (
    InvalidJsonFloat,
    JsonFloat,
    TQuantityValue,
)
from allotropy.exceptions import AllotropeConversionError, AllotropyParserError

PrimitiveValue = str | int | float


def str_to_bool(value: str) -> bool:
    return value.lower() in ("yes", "y", "true", "t", "1")


def try_int(value: str | None, value_name: str) -> int:
    try:
        return int(assert_not_none(value, value_name))
    except ValueError as e:
        msg = f"Invalid integer string: '{value}'."
        raise AllotropeConversionError(msg) from e


def try_int_or_none(value: str | None) -> int | None:
    try:
        return int(value or "")
    except ValueError:
        return None


def _try_float(value: str) -> float:
    # NOTE: this will convert a string with commas for thousands into a decimal, potentially introducing
    # an unexpected error, e.g. one thousand represented as 1,000 would get converted to 1.0
    # However, numbers are not usually represented like this in scientific output (we have no example of it)
    return float(value.replace(",", "."))


def try_float(value: str, value_name: str) -> float:
    assert_not_none(value, value_name)
    try:
        return _try_float(value)
    except ValueError as e:
        msg = f"Invalid float string: '{value}'."
        raise AllotropeConversionError(msg) from e


def try_float_or_none(value: str | float | None) -> float | None:
    try:
        return _try_float(str(value))
    except ValueError:
        return None


def try_nan_float_or_none(value: str | float | None) -> JsonFloat | None:
    float_value = try_float_or_none(value) if isinstance(value, str) else value
    if float_value is None:
        return None
    return InvalidJsonFloat.NaN if math.isnan(float_value) else float_value


def try_non_nan_float_or_none(value: str | float | None) -> float | None:
    float_value = try_float_or_none(value)
    return None if float_value is None or math.isnan(float_value) else float_value


def try_non_nan_float(value: str) -> float:
    float_value = try_non_nan_float_or_none(value)
    if float_value is None:
        msg = f"Invalid non nan float string: '{value}'."
        raise AllotropeConversionError(msg)
    return float_value


def try_int_or_nan(value: str | int | None) -> int | InvalidJsonFloat:
    if isinstance(value, int):
        return value
    float_value = try_non_nan_float_or_none(value)
    return InvalidJsonFloat.NaN if float_value is None else int(float_value)


def try_float_or_nan(value: str | float | None) -> JsonFloat:
    float_value = try_non_nan_float_or_none(value)
    return InvalidJsonFloat.NaN if float_value is None else float_value


def natural_sort_key(key: str) -> list[str]:
    """Returns a sort key that treats numeric substrings as parsed integers for comparison."""
    tokens = [token for token in re.split(r"(\d+)", key) if token]
    return [
        f"{int(token):>10}" if token.isdecimal() else token.lower() for token in tokens
    ]


QuantityType = TypeVar("QuantityType", bound=TQuantityValue)


def quantity_or_none(
    value_cls: type[QuantityType],
    value: JsonFloat | list[JsonFloat] | list[int] | None,
    index: int | None = None,
) -> QuantityType | None:
    if value is None:
        return None
    if isinstance(value, list):
        return value_cls(value=value[assert_not_none(index, msg="Cannot provide list to quantity_or_none without index")])  # type: ignore[call-arg]
    # Typing does not know that all subclasses of TQuantityValue have default value for unit set.
    return value_cls(value=value)  # type: ignore[call-arg]


T = TypeVar("T")


def assert_not_none(
    value: T | None, name: str | None = None, msg: str | None = None
) -> T:
    if value is None:
        msg = msg or f"Expected non-null value{f' for {name}' if name else ''}."
        raise AllotropeConversionError(msg)
    return value


def df_to_series(
    df: pd.DataFrame,
    msg: str,
) -> pd.Series[Any]:
    n_rows, _ = df.shape
    if n_rows == 1:
        return pd.Series(df.iloc[0], index=df.columns)
    raise AllotropyParserError(msg)


def assert_df_column(df: pd.DataFrame, column: str) -> pd.Series[Any]:
    df_column = df.get(column)
    if df_column is None:
        msg = f"Unable to find column '{column}'"
        raise AllotropeConversionError(msg)
    return pd.Series(df_column)


def assert_not_empty_df(df: pd.DataFrame, msg: str) -> pd.DataFrame:
    if df.empty:
        raise AllotropeConversionError(msg)
    return df


def assert_value_from_df(df: pd.DataFrame, key: str) -> Any:
    try:
        return df[key]
    except KeyError as e:
        msg = f"Unable to find key '{key}' in dataframe headers: {df.columns.tolist()}"
        raise AllotropeConversionError(msg) from e


def num_to_chars(n: int) -> str:
    d, m = divmod(n, 26)  # 26 is the number of ASCII letters
    return "" if n < 0 else num_to_chars(d - 1) + chr(m + 65)  # chr(65) = 'A'


def str_or_none(value: Any) -> str | None:
    return None if value is None else str(value)



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
import re

import pandas as pd

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ExperimentType
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_contents import (
    DesignQuantstudioContents,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    Referenceable,
)
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_df_column,
    assert_not_empty_df,
    assert_not_none,
    try_float,
    try_int,
)

SAMPLE_ROLE_TYPES_MAP = {
    "NTC": "negative control sample role",
    "STANDARD": "standard sample role",
    "UNKNOWN": "unknown sample role",
    "POSITIVE CONTROL": "positive control sample role",
    "IPC": "reference DNA control sample role",
    "BLOCKED_IPC": "DNA amplification control sample role",
    "POSITIVE_1/1": "homozygous control sample role",
    "POSITIVE_2/2": "homozygous control sample role",
    "POSITIVE_1/2": "heterozygous control sample role",
}


@dataclass(frozen=True)
class Header:
    measurement_time: str
    plate_well_count: int
    device_identifier: str
    model_number: str
    device_serial_number: str
    measurement_method_identifier: str
    pcr_detection_chemistry: str
    passive_reference_dye_setting: str | None
    barcode: str | None
    analyst: str | None
    experimental_data_identifier: str | None
    pcr_stage_number: int | None
    software_name: str | None
    software_version: str | None
    block_serial_number: str | None
    heated_cover_serial_number: str | None

    @staticmethod
    def create(header: SeriesData) -> Header:
        software_info = assert_not_none(
            re.match(
                "(.*) v(.+)",
                header[str, "Software Name and Version"],
            )
        )

        stage_number_raw = header.get(str, "PCR Stage/Step Number", "")
        stage_number = re.match(r"Stage (\d+)", stage_number_raw)
        pcr_stage_number = None if stage_number is None else int(stage_number.group(1))

        return Header(
            measurement_time=header[
                str,
                ["Run End Date/Time", "Run End Data/Time"],
                "Unable to find measurement time.",
            ],
            plate_well_count=assert_not_none(
                try_int(
                    assert_not_none(
                        re.match(
                            "(96)|(384)",
                            header[str, "Block Type"],
                        ),
                        msg="Unable to find plate well count",
                    ).group(),
                    "plate well count",
                ),
                msg="Unable to interpret plate well count",
            ),
            device_identifier=header.get(str, "Instrument Name", NOT_APPLICABLE),
            model_number=header.get(str, "Instrument Type", NOT_APPLICABLE),
            device_serial_number=header.get(
                str, "Instrument Serial Number", NOT_APPLICABLE
            ),
            measurement_method_identifier=header[str, "Quantification Cycle Method"],
            pcr_detection_chemistry=header.get(str, "Chemistry", NOT_APPLICABLE),
            passive_reference_dye_setting=header.get(str, "Passive Reference"),
            barcode=header.get(str, "Barcode"),
            analyst=header.get(str, "Operator"),
            experimental_data_identifier=header.get(str, "Experiment Name"),
            block_serial_number=header.get(str, "Block Serial Number"),
            heated_cover_serial_number=header.get(str, "Heated Cover Serial Number"),
            pcr_stage_number=pcr_stage_number,
            software_name=software_info.group(1),
            software_version=software_info.group(2),
        )


@dataclass
class WellItem(Referenceable):
    identifier: int
    target_dna_description: str
    sample_identifier: str
    reporter_dye_setting: str | None
    well_location_identifier: str | None
    quencher_dye_setting: str | None
    sample_role_type: str | None
    amplification_data: AmplificationData
    result: Result
    melt_curve_data: MeltCurveData | None = None

    @staticmethod
    def create(
        contents: DesignQuantstudioContents,
        data: SeriesData,
        experiment_type: ExperimentType,
    ) -> WellItem:
        identifier = data[int, "Well"]

        target_dna_description = data[
            str,
            "Target",
            f"Unable to find target dna description for well {identifier}",
        ]
        well_position = data[
            str,
            "Well Position",
            f"Unable to find well position for Well '{identifier}'.",
        ]

        amp_data = contents.get_non_empty_sheet("Amplification Data")
        melt_curve_data = contents.get_non_empty_sheet_or_none("Melt Curve Raw")

        return WellItem(
            uuid=random_uuid_str(),
            identifier=identifier,
            target_dna_description=target_dna_description,
            sample_identifier=data.get(str, "Sample", well_position),
            reporter_dye_setting=data.get(str, "Reporter"),
            well_location_identifier=well_position,
            quencher_dye_setting=data.get(str, "Quencher"),
            sample_role_type=SAMPLE_ROLE_TYPES_MAP.get(
                data.get(str, "Task", "__INVALID_KEY__")
            ),
            amplification_data=AmplificationData.create(
                amp_data, identifier, target_dna_description
            ),
            melt_curve_data=(
                None
                if melt_curve_data is None
                else MeltCurveData.create(
                    melt_curve_data, identifier, target_dna_description
                )
            ),
            result=Result.create(data, identifier, experiment_type),
        )


@dataclass
class Well:
    identifier: int
    items: dict[str, WellItem]
    multicomponent_data: MulticomponentData | None = None

    def get_well_item(self, target: str) -> WellItem:
        return assert_not_none(
            self.items.get(target),
            msg=f"Unable to find target DNA '{target}' for well {self.identifier}.",
        )

    @staticmethod
    def create(
        contents: DesignQuantstudioContents,
        header: Header,
        well_data: pd.DataFrame,
        identifier: int,
        experiment_type: ExperimentType,
    ) -> Well:
        well_items = {
            SeriesData(item_data)[str, "Target"]: WellItem.create(
                contents, SeriesData(item_data), experiment_type
            )
            for _, item_data in well_data.iterrows()
        }
        multi_data = contents.get_non_empty_sheet_or_none("Multicomponent")
        return Well(
            identifier=identifier,
            items=well_items,
            multicomponent_data=(
                None
                if multi_data is None
                else MulticomponentData.create(header, multi_data, identifier)
            ),
        )


@dataclass(frozen=True)
class WellList:
    wells: list[Well]

    def get_well_items(self) -> list[WellItem]:
        wells: list[WellItem] = []
        for well in self.wells:
            wells += well.items.values()
        return wells

    def __iter__(self) -> Iterator[Well]:
        return iter(self.wells)

    @staticmethod
    def get_well_result_data(
        contents: DesignQuantstudioContents, experiment_type: ExperimentType
    ) -> pd.DataFrame:
        data_sheet = (
            "Standard Curve Result"
            if experiment_type == ExperimentType.standard_curve_qPCR_experiment
            else "Results"
        )

        data = contents.get_non_empty_sheet(data_sheet)

        def add_data(extra_data: pd.DataFrame, columns: list[str]) -> None:
            data[columns] = None
            for _, row in extra_data.iterrows():
                sample_cond = data["Sample"] == row["Sample"]
                target_cond = data["Target"] == row["Target"]
                data.loc[sample_cond & target_cond, columns] = row[columns].to_list()

        if experiment_type == ExperimentType.relative_standard_curve_qPCR_experiment:
            add_data(
                extra_data=contents.get_non_empty_sheet("Replicate Group Result"),
                columns=[
                    "Cq SE",
                ],
            )

            add_data(
                extra_data=contents.get_non_empty_sheet("RQ Replicate Group Result"),
                columns=[
                    "EqCq Mean",
                    "Adjusted EqCq Mean",
                    "Delta EqCq Mean",
                    "Delta EqCq SD",
                    "Delta EqCq SE",
                    "Delta Delta EqCq",
                    "Rq",
                    "Rq Min",
                    "Rq Max",
                ],
            )
        elif experiment_type == ExperimentType.presence_absence_qPCR_experiment:
            add_data(
                extra_data=contents.get_non_empty_sheet("Target Call"),
                columns=[
                    "Call",
                ],
            )
        elif experiment_type == ExperimentType.genotyping_qPCR_experiment:
            genotyping_result = contents.get_non_empty_sheet("Genotyping Result")

            # The genotyping result data does not contain a target column
            # it can be constructed concatenating SNP assay column and the strings Allele 1/2
            rows = []
            for idx, row in genotyping_result.iterrows():
                snp_assay = assert_not_none(
                    row.get("SNP Assay"),
                    msg=f"Unable to get SNP Assay from Genotyping Result row '{idx}'.",
                )
                for allele in ["Allele 1", "Allele 2"]:
                    new_row = row.copy()
                    new_row["Target"] = f"{snp_assay}-{allele}"
                    rows.append(new_row)

            add_data(
                extra_data=pd.DataFrame(rows).reset_index(drop=True),
                columns=[
                    "Call",
                ],
            )

        return data

    @staticmethod
    def create(
        contents: DesignQuantstudioContents,
        header: Header,
        experiment_type: ExperimentType,
    ) -> WellList:
        results_data = WellList.get_well_result_data(contents, experiment_type)
        assert_df_column(results_data, "Well")

        return WellList(
            wells=[
                Well.create(
                    contents,
                    header,
                    well_data,
                    try_int(str(identifier), "well identifier"),
                    experiment_type,
                )
                for identifier, well_data in results_data.groupby("Well")
            ]
        )


@dataclass(frozen=True)
class AmplificationData:
    total_cycle_number_setting: float
    cycle: list[float]
    rn: list[float | None]
    delta_rn: list[float | None]

    @staticmethod
    def create(
        amplification_data: pd.DataFrame,
        well_item_id: int,
        target_dna_description: str,
    ) -> AmplificationData:
        well_data = assert_not_empty_df(
            amplification_data[
                assert_df_column(amplification_data, "Well") == well_item_id
            ],
            msg=f"Unable to find amplification data for well {well_item_id}.",
        )

        target_data = assert_not_empty_df(
            well_data[assert_df_column(well_data, "Target") == target_dna_description],
            msg=f"Unable to find amplification data for target '{target_dna_description}' in well {well_item_id}.",
        )

        cycle_number = assert_df_column(target_data, "Cycle Number")
        return AmplificationData(
            total_cycle_number_setting=try_float(
                str(cycle_number.max()), "Cycle Number"
            ),
            cycle=cycle_number.tolist(),
            rn=assert_df_column(target_data, "Rn").tolist(),
            delta_rn=assert_df_column(target_data, "dRn").tolist(),
        )


@dataclass(frozen=True)
class MulticomponentData:
    cycle: list[float]
    columns: dict[str, list[float | None]]

    def get_column(self, name: str) -> list[float | None]:
        return assert_not_none(
            self.columns.get(name),
            msg=f"Unable to obtain '{name}' from multicomponent data.",
        )

    @staticmethod
    def create(header: Header, data: pd.DataFrame, well_id: int) -> MulticomponentData:
        well_data = assert_not_empty_df(
            data[assert_df_column(data, "Well") == well_id],
            msg=f"Unable to find multi component data for well {well_id}.",
        )

        stage_number = well_data.get("Stage Number")
        stage_data = (
            well_data
            if header.pcr_stage_number is None or stage_number is None
            else assert_not_empty_df(
                well_data[stage_number == header.pcr_stage_number],  # type: ignore[arg-type]
                msg=f"Unable to find multi component data for stage {header.pcr_stage_number}.",
            )
        )

        return MulticomponentData(
            cycle=assert_df_column(stage_data, "Cycle Number").tolist(),
            columns={
                str(name): stage_data[name].tolist()
                for name in stage_data
                if name
                not in [
                    "Well",
                    "Cycle Number",
                    "Well Position",
                    "Stage Number",
                    "Step Number",
                ]
            },
        )


@dataclass(frozen=True)
class MeltCurveData:
    target: str
    temperature: list[float]
    fluorescence: list[float | None]
    derivative: list[float | None]

    @staticmethod
    def create(
        data: pd.DataFrame, well_id: int, target_dna_description: str
    ) -> MeltCurveData:
        well_data = assert_not_empty_df(
            data[assert_df_column(data, "Well") == well_id],
            msg=f"Unable to find melt curve data for well {well_id}.",
        )

        target_data = assert_not_empty_df(
            well_data[assert_df_column(well_data, "Target") == target_dna_description],
            msg=f"Unable to find melt curve data for target '{target_dna_description}' in well {well_id} .",
        )

        return MeltCurveData(
            target=target_dna_description,
            temperature=assert_df_column(target_data, "Temperature").tolist(),
            fluorescence=assert_df_column(target_data, "Fluorescence").tolist(),
            derivative=assert_df_column(target_data, "Derivative").tolist(),
        )


@dataclass(frozen=True)
class Result:
    cycle_threshold_value_setting: float
    cycle_threshold_result: float | None
    automatic_cycle_threshold_enabled_setting: bool | None
    automatic_baseline_determination_enabled_setting: bool | None
    normalized_reporter_result: float | None
    baseline_corrected_reporter_result: float | None
    baseline_determination_start_cycle_setting: float | None
    baseline_determination_end_cycle_setting: float | None
    genotyping_determination_result: str | None
    genotyping_determination_method_setting: float | None
    quantity: float | None
    quantity_mean: float | None
    quantity_sd: float | None
    ct_mean: float | None
    eq_ct_mean: float | None
    adj_eq_ct_mean: float | None
    ct_sd: float | None
    ct_se: float | None
    delta_ct_mean: float | None
    delta_ct_se: float | None
    delta_ct_sd: float | None
    delta_delta_ct: float | None
    rq: float | None
    rq_min: float | None
    rq_max: float | None
    rn_mean: float | None
    rn_sd: float | None
    y_intercept: float | None
    r_squared: float | None
    slope: float | None
    efficiency: float | None

    @staticmethod
    def get_reference_sample(contents: DesignQuantstudioContents) -> str:
        data = contents.get_non_empty_sheet("RQ Replicate Group Result")
        reference_data = data[assert_df_column(data, "Rq") == 1]
        reference_sample_array = assert_df_column(reference_data, "Sample").unique()

        if reference_sample_array.size != 1:
            msg = "Unable to infer reference sample, expecting a single row in sheet 'RQ Replicate Group Result' to have Rq == 1."
            raise AllotropeConversionError(msg)

        return str(reference_sample_array[0])

    @staticmethod
    def get_reference_target(contents: DesignQuantstudioContents) -> str | None:
        data = contents.get_non_empty_sheet("RQ Replicate Group Result")

        possible_ref_targets = set.intersection(
            *[
                set(
                    assert_df_column(
                        sample_data[assert_df_column(sample_data, "Rq").isnull()],
                        "Target",
                    ).tolist()
                )
                for _, sample_data in data.groupby("Sample")
            ]
        )

        if len(possible_ref_targets) == 0:
            return None

        if len(possible_ref_targets) == 1:
            return str(possible_ref_targets.pop())

        msg = "Unable to infer reference target, expecting a single unique value for Target in sheet 'RQ Replicate Group Result' where Rq is empty."
        raise AllotropeConversionError(msg)

    @staticmethod
    def create(
        target_data: SeriesData,
        well_item_id: int,
        experiment_type: ExperimentType,
    ) -> Result:
        genotyping_determination_result = (
            target_data.get(str, "Call")
            if experiment_type
            in (
                ExperimentType.presence_absence_qPCR_experiment,
                ExperimentType.genotyping_qPCR_experiment,
            )
            else None
        )

        genotyping_determination_method_setting = (
            target_data.get(float, "Threshold")
            if experiment_type
            in (
                ExperimentType.presence_absence_qPCR_experiment,
                ExperimentType.genotyping_qPCR_experiment,
            )
            else None
        )

        return Result(
            cycle_threshold_value_setting=target_data[
                float,
                "Threshold",
                f"Unable to find cycle threshold value setting for well {well_item_id}",
            ],
            cycle_threshold_result=target_data.get(float, "Cq"),
            automatic_cycle_threshold_enabled_setting=target_data.get(
                bool, "Auto Threshold", None
            ),
            automatic_baseline_determination_enabled_setting=target_data.get(
                bool, "Auto Baseline", None
            ),
            normalized_reporter_result=target_data.get(float, "Rn"),
            baseline_corrected_reporter_result=target_data.get(float, "Delta Rn"),
            baseline_determination_start_cycle_setting=target_data.get(
                float, "Baseline Start"
            ),
            baseline_determination_end_cycle_setting=target_data.get(
                float, "Baseline End"
            ),
            genotyping_determination_result=genotyping_determination_result,
            genotyping_determination_method_setting=genotyping_determination_method_setting,
            quantity=target_data.get(float, "Quantity"),
            quantity_mean=target_data.get(float, "Quantity Mean"),
            quantity_sd=target_data.get(float, "Quantity SD"),
            ct_mean=target_data.get(float, "Cq Mean"),
            eq_ct_mean=target_data.get(float, "EqCq Mean"),
            adj_eq_ct_mean=target_data.get(float, "Adjusted EqCq Mean"),
            ct_sd=target_data.get(float, "Cq SD"),
            ct_se=target_data.get(float, "Cq SE"),
            delta_ct_mean=target_data.get(float, "Delta EqCq Mean"),
            delta_ct_se=target_data.get(float, "Delta EqCq SE"),
            delta_ct_sd=target_data.get(float, "Delta EqCq SD"),
            delta_delta_ct=target_data.get(float, "Delta Delta EqCq"),
            rq=target_data.get(float, "Rq"),
            rq_min=target_data.get(float, "Rq Min"),
            rq_max=target_data.get(float, "Rq Max"),
            rn_mean=target_data.get(float, "Rn Mean"),
            rn_sd=target_data.get(float, "Rn SD"),
            y_intercept=target_data.get(float, "Y-Intercept"),
            r_squared=target_data.get(float, "R2"),
            slope=target_data.get(float, "Slope"),
            efficiency=target_data.get(float, "Efficiency"),
        )


@dataclass(frozen=True)
class Data:
    header: Header
    wells: WellList
    experiment_type: ExperimentType
    calculated_documents: list[CalculatedDocument]
    reference_target: str | None
    reference_sample: str | None

    @staticmethod
    def get_experiment_type(contents: DesignQuantstudioContents) -> ExperimentType:
        experiment_type_to_expected_sheets = {
            ExperimentType.standard_curve_qPCR_experiment: ["Standard Curve Result"],
            ExperimentType.relative_standard_curve_qPCR_experiment: [
                "RQ Replicate Group Result"
            ],
            ExperimentType.genotyping_qPCR_experiment: ["Genotyping Result"],
            ExperimentType.melt_curve_qPCR_experiment: [
                "Melt Curve Raw",
                "Melt Curve Result",
            ],
            ExperimentType.presence_absence_qPCR_experiment: [
                "Sample Call",
                "Well Call",
                "Target Call",
                "Control Status",
            ],
        }

        possible_experiment_types = {
            experiment_type
            for experiment_type, expected_sheets in experiment_type_to_expected_sheets.items()
            if all(contents.has_sheet(sheet_name) for sheet_name in expected_sheets)
        }

        if len(possible_experiment_types) == 1:
            return possible_experiment_types.pop()

        msg = f"Unable to infer experiment type from sheets in the input, expected exactly one set of sheets from: {list(experiment_type_to_expected_sheets.values())}, got {list(contents.data.keys())}"
        raise AllotropeConversionError(msg)



from xml.etree import ElementTree

from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.values import try_float_or_nan, try_float_or_none


def get_element_from_xml(
    xml_object: ElementTree.Element, tag_name: str, tag_name_2: str | None = None
) -> ElementTree.Element:
    if tag_name_2 is not None:
        tag_finder = tag_name + "/" + tag_name_2
        xml_element = xml_object.find(tag_finder)
    else:
        tag_finder = tag_name
        xml_element = xml_object.find(tag_finder)
    if xml_element is not None:
        return xml_element
    else:
        msg = f"Unable to find '{tag_finder}' from xml."
        raise AllotropeConversionError(msg)


def get_val_from_xml(
    xml_object: ElementTree.Element, tag_name: str, tag_name_2: str | None = None
) -> str:
    return str(get_element_from_xml(xml_object, tag_name, tag_name_2).text)


def get_val_from_xml_or_none(
    xml_object: ElementTree.Element, tag_name: str, tag_name_2: str | None = None
) -> str | None:
    try:
        val_from_xml = get_element_from_xml(xml_object, tag_name, tag_name_2).text
        if val_from_xml is not None:
            return str(val_from_xml)
        else:
            return None
    except AllotropeConversionError:
        return None


def get_float_from_xml_or_nan(
    xml_object: ElementTree.Element, tag_name: str, tag_name_2: str | None = None
) -> JsonFloat:
    return try_float_or_nan(get_val_from_xml_or_none(xml_object, tag_name, tag_name_2))


def get_float_from_xml_or_none(
    xml_object: ElementTree.Element, tag_name: str, tag_name_2: str | None = None
) -> float | None:
    return try_float_or_none(get_val_from_xml_or_none(xml_object, tag_name, tag_name_2))


def get_attrib_from_xml(
    xml_object: ElementTree.Element,
    tag_name: str,
    attrib_name: str,
    tag_name_2: str | None = None,
) -> str:
    xml_element = get_element_from_xml(xml_object, tag_name, tag_name_2)
    try:
        attribute_val = xml_element.attrib[attrib_name]
        return attribute_val
    except KeyError as e:
        msg = f"Unable to find '{attrib_name}' in {xml_element.attrib}"
        raise AllotropeConversionError(msg) from e



from allotropy.allotrope.models.adm.multi_analyte_profiling.benchling._2024._01.multi_analyte_profiling import (
    AnalyteAggregateDocument,
    AnalyteDocumentItem,
    CalibrationAggregateDocument,
    CalibrationDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    MultiAnalyteProfilingAggregateDocument,
    MultiAnalyteProfilingDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicroliter,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TStatisticDatumRole,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.luminex_xponent.luminex_xponent_structure import (
    Data,
    Header,
    Measurement,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser

DEFAULT_SOFTWARE_NAME = "xPONENT"
DEFAULT_CONTAINER_TYPE = "well plate"
DEFAULT_DEVICE_TYPE = "multi analyte profiling analyzer"


class LuminexXponentParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Luminex xPONENT"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        lines = read_to_lines(named_file_contents)
        reader = CsvReader(lines)
        data = Data.create(reader)
        return self._get_model(named_file_contents.original_file_name, data)

    def _get_model(self, file_name: str, data: Data) -> Model:
        header = data.header
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/multi-analyte-profiling/BENCHLING/2024/01/multi-analyte-profiling.manifest",
            multi_analyte_profiling_aggregate_document=MultiAnalyteProfilingAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=header.model_number,
                    equipment_serial_number=header.equipment_serial_number,
                    calibration_aggregate_document=CalibrationAggregateDocument(
                        calibration_document=[
                            CalibrationDocumentItem(
                                calibration_name=calibration_item.name,
                                calibration_report=calibration_item.report,
                                calibration_time=self._get_date_time(
                                    calibration_item.time
                                ),
                            )
                            for calibration_item in data.calibration_data
                        ]
                    ),
                ),
                data_system_document=DataSystemDocument(
                    data_system_instance_identifier=header.data_system_instance_identifier,
                    file_name=file_name,
                    software_name=DEFAULT_SOFTWARE_NAME,
                    software_version=header.software_version,
                    ASM_converter_name=self.get_asm_converter_name(),
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                multi_analyte_profiling_document=[
                    MultiAnalyteProfilingDocumentItem(
                        analyst=header.analyst,
                        measurement_aggregate_document=MeasurementAggregateDocument(
                            analytical_method_identifier=header.analytical_method_identifier,
                            method_version=header.method_version,
                            experimental_data_identifier=header.experimental_data_identifier,
                            container_type=DEFAULT_CONTAINER_TYPE,
                            plate_well_count=TQuantityValueNumber(
                                value=header.plate_well_count
                            ),
                            measurement_document=[
                                self._get_measurement_document_item(
                                    measurement,
                                    header,
                                    data.minimum_bead_count_setting,
                                )
                            ],
                        ),
                    )
                    for measurement in data.measurement_list.measurements
                ],
            ),
        )

    def _get_measurement_document_item(
        self,
        measurement: Measurement,
        header_data: Header,
        minimum_bead_count_setting: float,
    ) -> MeasurementDocumentItem:
        error_aggregate_document = None
        if measurement.errors:
            error_aggregate_document = ErrorAggregateDocument(
                error_document=[
                    ErrorDocumentItem(error=error) for error in measurement.errors
                ]
            )

        return MeasurementDocumentItem(
            measurement_identifier=random_uuid_str(),
            measurement_time=self._get_date_time(header_data.measurement_time),
            sample_document=SampleDocument(
                sample_identifier=measurement.sample_identifier,
                location_identifier=measurement.location_identifier,
            ),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        device_type=DEFAULT_DEVICE_TYPE,
                        sample_volume_setting=TQuantityValueMicroliter(
                            value=header_data.sample_volume_setting
                        ),
                        dilution_factor_setting=TQuantityValueUnitless(
                            value=measurement.dilution_factor_setting
                        ),
                        detector_gain_setting=header_data.detector_gain_setting,
                        minimum_assay_bead_count_setting=TQuantityValueNumber(
                            value=minimum_bead_count_setting
                        ),
                    )
                ]
            ),
            assay_bead_count=TQuantityValueNumber(value=measurement.assay_bead_count),
            analyte_aggregate_document=AnalyteAggregateDocument(
                analyte_document=[
                    AnalyteDocumentItem(
                        analyte_identifier=random_uuid_str(),
                        analyte_name=analyte.analyte_name,
                        assay_bead_identifier=analyte.assay_bead_identifier,
                        assay_bead_count=TQuantityValueNumber(
                            value=analyte.assay_bead_count
                        ),
                        fluorescence=TQuantityValueRelativeFluorescenceUnit(
                            value=analyte.fluorescence,
                            has_statistic_datum_role=TStatisticDatumRole.median_role,
                        ),
                    )
                    for analyte in measurement.analytes
                ]
            ),
            error_aggregate_document=error_aggregate_document,
        )



import unicodedata

import pandas as pd

from allotropy.allotrope.pandas_util import read_csv
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import determine_encoding


class ViCellBluReader:
    @classmethod
    def read(cls, named_file_contents: NamedFileContents) -> pd.DataFrame:
        contents = named_file_contents.contents.read()
        encoding = (
            determine_encoding(contents, named_file_contents.encoding)
            if isinstance(contents, bytes)
            else None
        )
        named_file_contents.contents.seek(0)
        return read_csv(
            named_file_contents.contents, index_col=False, encoding=encoding
        ).rename(columns=cls._normalize_column_names)

    @classmethod
    def _normalize_column_names(cls, col: str) -> str:
        """Normalize column names that look the same but have mismatching characters

        Some column names in the instrument files include two different `µ` characters that look the same,
        the Micro Sign (U+00B5) and the Greek Small letter mu (U+03BC), this mapper uses the unicode `NFKC`
        normalization to ensure consistency when retrieving values from those columns.
        """
        return unicodedata.normalize("NFKC", col)
